#include "../lib/mysql.hpp"


marcelb::mysql::MySQL::MySQL(const string _path, const string _username, const string _password, const string _db, const uint _available, const periodical_engine _engine_type) {
   path = _path;
   username = _username;
   password = _password;
   db = _db;
   available = _available > 0 ? _available : 1;
   engine_type = _engine_type;

   drv = get_mysql_driver_instance();

   if (engine_type == periodical_engine::internal) {
      periodic_engin = async(launch::async, [&](){
         while (run_engin) {
            usleep(MYSQL_PERIODIC_INTERNAL_TIME*1000);
            periodic_maintenance();
         }
         return;
      });  
   }
}


Connection* marcelb::mysql::MySQL::create_con() {
   uint trys = 0;
   bool status = true;
   Connection* new_con = NULL;

   while (reconTrys == unlimited ? status : (trys <= reconTrys && status)) {
      try {
         Connection* con_can = drv->connect(path, username, password);
         status = !con_can->isValid();
         if (!status) {
            new_con = con_can;
         }
         else if (!con_can->isClosed()) {
            disconnect_one(con_can);
         }
      }
      catch (const SQLException &error) {
         cout << error.what() << endl;
         usleep(reconnectSleep);
         reconTrys == unlimited ? trys : trys++;
      }         
   }

   return new_con;
}

bool marcelb::mysql::MySQL::disconnect() {
   io.lock();
   bool status = true;

   for (uint i=0; i<con.size(); i++) {
      status = disconnect_one(con[i]) ;
   }

   io.unlock();
   return status;
}

bool marcelb::mysql::MySQL::disconnect_one(Connection* con_ptr) {
   bool status = !con_ptr->isClosed();

   if (status) {
      try {
         con_ptr->close();
         status = !con_ptr->isClosed();
      } 
      catch (const SQLException &error) {
         cout << error.what() << endl;
         status = true;
      }
   }

   else {
      status = false; // već je zatvorena
   }

   delete con_ptr;
   return status;
}

bool marcelb::mysql::MySQL::open_one(Connection* con_ptr) {
   bool status = true; // ako true greška je
   uint trys = 0;

   while (reconTrys == unlimited ? status : (trys <= reconTrys && status)) {
      try {
         if (con_ptr->isValid()) {
            con_ptr->setSchema(db);
            status = false;
         }
         else {
            break;
         }
      }
      catch (const SQLException &error) {
         cout << error.what() << endl;
         usleep(reconnectSleep);
         reconTrys == unlimited ? trys : trys++;
      }
   }

   return status;
}

/**
 * Broj pokušaja usljed povezivanja s bazom od 1 do unlimited;
*/

void marcelb::mysql::MySQL::reconnectTrys(const uint _trys) {
   io.lock();
   reconTrys = _trys;
   io.unlock();
}

Connection* marcelb::mysql::MySQL::shift_con() {
   while (true) {
      while(con.size()) {
         io.lock();
         Connection* con_ptr = con[0];
         con.pop_front();
         if (con_ptr->isValid()) {
            io.unlock();
            return con_ptr;
         } 
         io.unlock();
      }
      usleep(1000);
   }
}

marcelb::mysql::MySQL::~MySQL() {
   if (engine_type == periodical_engine::internal) {
      run_engin = false;
      periodic_engin.get();
   } else {

   }
   
   disconnect();
}


void marcelb::mysql::MySQL::periodic_maintenance() {
   while (available>con.size() && run_engin) {
      try {
         Connection* new_con_ptr = create_con();
         if (!db.empty()) {
            if (open_one(new_con_ptr)) {
               throw string("[ERROR] Unable to open database " + db);
            }
         }
         io.lock();
         con.push_back(new_con_ptr);       
         io.unlock();
      } catch (const SQLException except) {
         cout << except.what() << endl;
      } catch (const string except) {
         cout << except << endl;
      }     
   }

   for (int i=0; i<con.size() && run_engin; i++) {
      if (!con[i]->isValid()) {
         io.lock();
         con.erase(con.begin()+i);
         io.unlock();
         i--;
      }
   }
}